package org.artifactory.repo.service.trash;

import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.properties.PropertiesService;
import org.artifactory.api.security.SecurityService;
import org.artifactory.fs.FileInfo;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoRepoPath;
import org.artifactory.repo.db.DbLocalRepo;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.artifactory.storage.db.fs.session.SqlStorageSession;
import org.artifactory.storage.fs.lock.SessionLockEntry;
import org.artifactory.storage.fs.session.StorageSessionHolder;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.Calendar;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Shay Bagants
 */
@Test
public class TrashServiceImplTest extends ArtifactoryHomeBoundTest {
    @Mock
    private InternalRepositoryService repoService;
    @Mock
    private SecurityService securityService;
    @Mock
    private InternalBinaryService binaryStore;
    @Mock
    private CentralConfigService configService;
    @Mock
    private PropertiesService propertiesService;
    @Mock
    private org.artifactory.storage.fs.service.PropertiesService storagePropsService;
    @Mock
    private NonTrashableItems nonTrashableItems;
    TrashServiceImpl trashService;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.initMocks(this);
        trashService = new TrashServiceImpl(repoService, securityService, binaryStore, configService, propertiesService,
                storagePropsService, nonTrashableItems);
    }

    @AfterMethod
    public void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
    }

    /**
     * The time that a file was stored in a trashcan is saved in the DB as millis as text.
     * The GC (trashcan cleanup phase) search for files that are in the trashcan for more than x days, and delete these.
     * The date (timestamp) comparision is done in the DB, on the text values, therefore, it is mandatory that all
     * timestamps will be at the same length (same number of digits)
     *
     * We decided that the min allowed date (oldest time of file in trashcan) is 2003 January 1
     */
    @Test(dataProvider = "provideInvalidTimestamps", expectedExceptions = IllegalArgumentException.class,
            expectedExceptionsMessageRegExp = "Illegal retention date. Retention period goes back too far.")
    public void testValidateRetentionPeriodTimestampInvalidTimestamps(long timetamp) {
        trashService.validateRetentionPeriodTimestamp(timetamp);
    }

    @DataProvider
    public static Object[][] provideInvalidTimestamps() {
        Object[][] objects = new Object[22][];
        Calendar cal = Calendar.getInstance();
        int i = 0;
        for (int year = 1990; year < 2002; year++) {
            cal.set(year, 01, 01);
            objects[i] = new Object[]{cal.getTime().toInstant().toEpochMilli()};
            i++;
        }
        return objects;
    }

    @Test(dataProvider = "provideValidTimestamps")
    public void testValidateRetentionPeriodTimestamp(long timstamp) {
        trashService.validateRetentionPeriodTimestamp(timstamp);
    }

    @Test
    public void testUndeployIfAncestorFilesInTrash() {
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        when(repoService.getFileInfo(any())).thenReturn(mock(FileInfo.class));
        when(repoService.undeploy(any())).thenReturn(mock(BasicStatusHolder.class));
        DbLocalRepo dbLocalRepoMock = mock(DbLocalRepo.class);
        RepoRepoPath<Repo> targetRepoRepoPath =
                new RepoRepoPath(dbLocalRepoMock, new RepoPathImpl("auto-trashcan", "/parent/foo.bar"));
        when(repoService.getRepoRepoPath(any())).thenReturn(targetRepoRepoPath);
        when(context.beanForType(InternalRepositoryService.class)).thenReturn(repoService);
        ArtifactoryContextThreadBinder.bind(context);

        SqlStorageSession sqlStorageSessionMock = mock(SqlStorageSession.class);
        SessionLockEntry sessionLockEntryMock = mock(SessionLockEntry.class);
        when(sqlStorageSessionMock.getLockEntry(any())).thenReturn(sessionLockEntryMock);
        StorageSessionHolder.setSession(sqlStorageSessionMock);

        trashService.undeployIfAncestorFilesInTrash(new RepoPathImpl("auto-trashcan", "/parent/foo.bar"));

        verify(sessionLockEntryMock, times(2)).setWriteFsItem(any());
        verify(dbLocalRepoMock, times(1)).createOrGetFolder(any());
    }

    @DataProvider
    public static Object[][] provideValidTimestamps() {
        Object[][] objects = new Object[27][];
        Calendar cal = Calendar.getInstance();
        int i = 0;
        for (int year = 2003; year < 2030; year++) {
            cal.set(year, 01, 02);
            objects[i] = new Object[]{cal.getTime().toInstant().toEpochMilli()};
            i++;
        }
        return objects;
    }
}