/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.locks.service;

import lombok.SneakyThrows;
import org.apache.commons.compress.utils.Sets;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.dao.locks.DbDistributeLocksDao;
import org.artifactory.storage.db.locks.DbUnlockSupplier;
import org.artifactory.storage.db.locks.LockInfo;
import org.artifactory.storage.fs.lock.ThreadDumpUtils;
import org.artifactory.util.CollectionUtils;
import org.artifactory.version.CompoundVersionDetails;
import org.jfrog.common.config.diff.DataDiff;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import static org.artifactory.common.ConstantValues.hazelcastMaxLockLeaseTime;
import static org.artifactory.storage.db.locks.service.HaDbLockHelper.getLockInfo;
import static org.jfrog.storage.util.DbUtils.tableExists;

/**
 * @author gidis
 */
@Service
@Reloadable(beanClass = DbLocksService.class, initAfter = DbService.class)
public class DbLocksServiceImpl implements DbLocksService {
    private static final Logger log = LoggerFactory.getLogger(DbLocksServiceImpl.class);

    private final DbDistributeLocksDao dbDistributeLocksDao;
    private final Map<String, LockInfo> localLocks = new ConcurrentHashMap<>();
    private final Set<Pair<String, String>> stuckLocks = Collections.newSetFromMap(new ConcurrentHashMap<>());
    private final int maxWaitTimeMs;
    private final int initialWaitMs;
    private final int maxRetryForDeletions;
    private final int initialUnlockWaitMs;
    private final ArtifactoryDbProperties dbProperties;

    @Autowired
    public DbLocksServiceImpl(DbDistributeLocksDao dbDistributeLocksDao, ArtifactoryDbProperties dbProperties) {
        this.dbDistributeLocksDao = dbDistributeLocksDao;
        this.dbProperties = dbProperties;
        this.maxWaitTimeMs = ConstantValues.dbLockMaxWaitMs.getInt();
        this.initialWaitMs = ConstantValues.dbLockInitialWaitMs.getInt();
        this.maxRetryForDeletions = ConstantValues.dbLockMaxRetryForDeletions.getInt();
        this.initialUnlockWaitMs = ConstantValues.initialUnlockWaitMs.getInt();
    }

    private static void waitLimitedTime(long timeout, long sleepTime, long waitingTime) throws TimeoutException {
        if ((timeout > 0) && (waitingTime < timeout)) {
            long maxTimeTOWait = timeout - waitingTime;
            long timeToSleep = Math.min(sleepTime, maxTimeTOWait);
            sleep(timeToSleep);
        } else {
            throw new TimeoutException("Couldn't acquire lock for: " + timeout + " milliseconds");
        }
    }

    private static void sleep(long timeToSleep) {
        try {
            log.trace("Going to sleep for: {}", timeToSleep);
            Thread.sleep(timeToSleep);
        } catch (InterruptedException e) {
            log.warn("Interrupted exception occur while waiting to acquire lock.", e);
            Thread.currentThread().interrupt();
        }
    }

    private static String getLockInfoMsg(String category, String key) {
        return getLockInfoMsg(category, key, null);
    }

    private static String getLockInfoMsg(String category, String key, String owner) {
        StringBuilder builder = new StringBuilder();
        if (StringUtils.isNotBlank(category)) {
            builder.append("category: '").append(category).append("' ");
        }
        if (StringUtils.isNotBlank(key)) {
            builder.append("key: '").append(key).append("' ");
        }
        if (StringUtils.isNotBlank(owner)) {
            builder.append("owner: '").append(owner).append("' ");
        }
        return builder
                .append("thread id: '")
                .append(Thread.currentThread().getId())
                .append("' thread name: '")
                .append(Thread.currentThread().getName())
                .append("'")
                .toString();
    }

    @Override
    public void acquireLock(String category, String key, String owner, long timeout, TimeUnit timeUnit)
            throws TimeoutException {
        timeout = timeUnit.toMillis(timeout);
        long start = System.currentTimeMillis();
        String lockInfoMsg = getLockInfoMsg(category, key, owner);
        log.debug("Acquiring lock for: {}", lockInfoMsg);
        long startTime = System.currentTimeMillis();
        LockInfo lockInfo = new LockInfo(category, key, owner, Thread.currentThread().getId(),
                Thread.currentThread().getName(), start);
        boolean dbLocked = false;
        try {
            dbLocked = acquireLock(lockInfo, lockInfoMsg, timeout, start, startTime);
        } finally {
            final String localId = toLocalId(category, key);
            if (!dbLocked && lockInfo.equals(localLocks.get(localId))) {
                localLocks.remove(localId);
            }
        }
    }

    @SneakyThrows
    private boolean acquireLock(LockInfo lockInfo, String lockInfoMsg,
            long timeout, long start, long startTime) {
        long sleepTime = initialWaitMs;
        long waitingTime;
        final String category = lockInfo.getCategory();
        final String owner = lockInfo.getOwner();
        final String key = lockInfo.getKey();
        while (true) {
            try {
                //Assume key is not locked
                log.trace("Attempting to acquire lock for: {}", lockInfoMsg);
                // we update the time on the model before inserting into the db, because in case we wait for x time
                // until we were able to acquire the lock, we don't want its expiry to be shorter.
                lockInfo.setStartedTime(System.currentTimeMillis());
                String localKey = toLocalId(category, key);
                //put lock to localLocks(jvm) and after it validate in db
                LockInfo lockInfoInJvm = localLocks.putIfAbsent(localKey, lockInfo);
                if (lockInfoInJvm != null && lockInfoInJvm != lockInfo) {
                    validateLockIsNotMine(lockInfoInJvm, owner);
                } else if (dbDistributeLocksDao.tryToAcquireLock(lockInfo)) {
                    log.trace("Successfully acquired lock: '{}'.", lockInfo);
                    long timeTookToAcquireLock = System.currentTimeMillis() - start;
                    log.debug("Lock acquired in '{}' milliseconds.", timeTookToAcquireLock);
                    return true;
                } else {
                    LockInfo dbLockInfo = dbDistributeLocksDao.getLockInfo(category, key);
                    if (dbLockInfo != null) {
                        validateLockIsNotMine(dbLockInfo, owner);
                    }
                }
            } catch (SQLException e) {
                if (!locksTableExists(category, key, owner)) {
                    log.debug("distributed_locks table doesn't exist yet. Retrying");
                } else {
                    log.error("Exception occurred while trying to acquire lock", e);
                    throw new IllegalStateException("Exception occurred while trying to acquire lock: " +
                            lockInfoMsg, e);
                }
            } catch (Exception e) {
                log.error("Failed to acquire lock for: {}", lockInfo, e);
                throw new IllegalStateException("Failed to acquire lock for: " + lockInfoMsg, e);
            }
            // If we got here the lock exist but we don't own it
            long currentTime = System.currentTimeMillis();
            waitingTime = currentTime - startTime;
            waitLimitedTime(timeout, sleepTime, waitingTime);
            sleepTime = Math.min(sleepTime * 4, maxWaitTimeMs);
            log.trace("Waiting while trying to acquiring lock for: {}", lockInfoMsg);
        }
    }

    private void validateLockIsNotMine(@Nonnull LockInfo lockInfo, String owner) {
        if (isLockMine(lockInfo, owner)) {
            throw new IllegalStateException("Reentrant lock is not supported");
        }
    }

    private boolean isLockMine(@Nonnull LockInfo lockInfo, String owner) {
        return (lockInfo.getThreadId() == Thread.currentThread().getId() &&
                lockInfo.getOwner().equals(owner));
    }

    @Override
    public boolean isLocked(String category, String key) {
        String lockInfoMsg = getLockInfoMsg(category, key);
        log.debug("Checking if lock exist for: {}", lockInfoMsg);
        try {
            if (localLocks.containsKey(toLocalId(category, key))) {
                log.trace("Lock exists locally for: {}", lockInfoMsg);
                return true;
            }
            if (dbDistributeLocksDao.isLocked(category, key)) {
                log.trace("Lock exist for: {}", lockInfoMsg);
                return true;
            }
            log.trace("Lock doesn't exist for: {}", lockInfoMsg);
            return false;
        } catch (SQLException e) {
            log.debug("Failed to check if lock exist for:" + lockInfoMsg, e);
            throw new IllegalStateException("Failed to check if lock exist for:" + lockInfoMsg, e);
        }
    }


    @Override
    public boolean isLockedByMe(String category, String key) {
        LockInfo lockInfo = localLocks.get(toLocalId(category, key));
        return lockInfo != null && Thread.currentThread().getId() == lockInfo.getThreadId();
    }


    @Override
    public boolean unlock(String category, String key, String owner) {
        String lockInfoMsg = getLockInfoMsg(category, key, owner);
        log.debug("Attempting to release lock for: {}", lockInfoMsg);
        try {
            // Assume lock acquired by this thread and try to update lock
            log.trace("Attempting to release lock from cache for: {}", lockInfoMsg);
            LockInfo lockInfo = localLocks.get(toLocalId(category, key));
            if (lockInfo != null) {
                if (lockInfo.getThreadId() == Thread.currentThread().getId()) {
                    localLocks.remove(toLocalId(category, key));
                    log.trace("Successfully delete lock from cache for: {}", lockInfo);
                } else {
                    String msg = "Failed to release lock (inconsistent state) for: {}. Current thread is not the " +
                            "owner of the lock." + lockInfo;
                    StringBuilder msgBuilder = new StringBuilder().append(msg);
                    ThreadDumpUtils.builder()
                            .count(1)
                            .build()
                            .dumpThreads(msgBuilder);
                    throw new IllegalStateException(msgBuilder.toString());
                }
            }
            DbUnlockSupplier unlockSupplier = buildUnlockSupplier(category, key, owner);
            boolean dbLockRemoval = unlockInternal(unlockSupplier, category, key, owner, maxRetryForDeletions);
            if (dbLockRemoval) {
                log.trace("Successfully delete lock from cache for: {}", lockInfoMsg);
                return true;
            } else {
                log.error("Could not remove {} lock.", lockInfoMsg);
            }
        } catch (Exception e) {
            log.error("Failed to release lock for: {}. Exception message: {}", lockInfoMsg, e.getMessage());
            log.debug("Failed to release lock for: {}", lockInfoMsg, e);
            throw new IllegalStateException("Failed to release lock for: " + lockInfoMsg, e);
        }
        log.debug("Lock is either not exist or not owned by current owner for: {}", lockInfoMsg);
        return false;
    }

    @Override
    public boolean forceUnlock(String category, String key) {
        String lockInfoMsg = getLockInfoMsg(category, key);
        log.debug("Releasing lock for: {}", lockInfoMsg);
        try {
            LockInfo removed = localLocks.remove(toLocalId(category, key));
            DbUnlockSupplier forceUnlockSupplier = getForceUnlockSupplier(category, key);
            boolean dbLockRemoved = unlockInternal(forceUnlockSupplier, category, key, null, maxRetryForDeletions);
            return removed != null || dbLockRemoved;
        } catch (SQLException e) {
            throw new IllegalStateException("Failed to force lock release for: " + lockInfoMsg, e);
        }
    }

    @Override
    public void forceUnlockAllStuck() {
        log.debug("Releasing all stuck locks.");
        stuckLocks.forEach(
                (pair) -> {
                    try {
                        boolean isSuccess = forceUnlock(pair.getLeft(), pair.getRight());
                        log.warn(isSuccess ? "Force unlocked: {}, {}" : "Unable to force unlock: {}, {}",
                                pair.getLeft(), pair.getRight());
                    } catch (IllegalStateException e) {
                        log.warn("Unable to force unlock : {}, {}.", pair.getLeft(), pair.getRight(), e);
                    }
                }
        );
        stuckLocks.clear();
    }

    /**
     * The first startup may happen from multiple nodes in parallel. This node may have yielded for another node in the
     * schema creation phase, so we give the node a chance to create the locks table.
     */
    private boolean locksTableExists(String category, String key, String owner) {
        try (Connection con = dbDistributeLocksDao.getLocksConnection()) {
            return tableExists(con.getMetaData(), dbProperties.getDbType(), "distributed_locks");
        } catch (SQLException se) {
            throw new IllegalStateException(
                    "Failed to check if distributed_locks table exists while trying to lock: " +
                            getLockInfoMsg(category, key, owner), se);
        }
    }

    private long getMinAllowedTimeMillis(long current, long maxLockLeaseTimeMinutes) {
        return current - TimeUnit.MINUTES.toMillis(maxLockLeaseTimeMinutes);
    }

    /**
     * Unlock a combination of 'category', 'key' and optionally 'owner' from the DB with retries and short sleep between
     * each retry
     *
     * @param numOfRetries Num of times to retry if the first attempt failed
     * @throws SQLException if the lock can't be removed
     */
    private boolean unlockInternal(DbUnlockSupplier unlockSupplier, String category, String key, String owner,
            int numOfRetries) throws SQLException {
        // basically, we need retry when catching an exception only, however, although it does not makes sense
        //without DB cluster, we saw once that a lock was exists in Oracle DB, and the unlock operation returned false.
        boolean success = false;
        int sleepTime = initialUnlockWaitMs; // sleep time in ms
        while (numOfRetries >= 0) {
            try {
                success = unlockSupplier.unlock();
            } catch (SQLException e) {
                log.trace("SQL Error occurred while trying to delete lock.", e);
                if (numOfRetries == 0) {
                    stuckLocks.add(Pair.of(category, key));
                    throw e;
                }
            }
            if (success) {
                return true;
            }
            log.debug("Failed removing lock for {}", getLockInfo(category, key, owner));
            if (numOfRetries > 0) {
                // sleep, no more than half a second
                sleepTime = Math.min(sleepTime, 256);
                sleep(sleepTime);
                sleepTime = sleepTime * 4;
            }
            numOfRetries--;
        }
        return false;
    }

    @Override
    public int lockingMapSize(String category) {
        log.debug("Querying for category map size for: " + getLockInfo(category, null));
        try {
            int numberOfLocks = dbDistributeLocksDao.lockingMapSize(category);
            log.trace("Found " + numberOfLocks + " locks for:" + getLockInfo(category));
            return numberOfLocks;
        } catch (SQLException e) {
            String message = StringUtils.isNotBlank(e.getMessage()) ? e.getMessage() : "";
            throw new RuntimeException(
                    "Failed to query out number of locks for: " + getLockInfo(category) + ". " + message);
        }
    }

    @Override
    public Set<String> lockingMapKeySet(String category) {
        log.debug("Querying for category map size for: " + getLockInfo(category));
        try {
            Set<String> set = dbDistributeLocksDao.lockingMapKeySet(category);
            log.trace("Found the following locks " + set + "  for:" + getLockInfo(category));
            return set;
        } catch (SQLException e) {
            throw new RuntimeException("Failed to query out locks for:" + getLockInfo(category));
        }
    }

    @Override
    public void cleanOrphanLocks(List<String> serverIdToKeepLocks) {
        if (CollectionUtils.isNullOrEmpty(serverIdToKeepLocks)) {
            throw new IllegalStateException("List of servers to keep locks must not be empty.");
        }
        try {
            Set<LockInfo> orphanLocks = dbDistributeLocksDao.getAllLocksNotOwnedBy(serverIdToKeepLocks);
            log.debug("Attempting to remove {} orphan locks", orphanLocks.size());
            forceRemoveLocks(orphanLocks, false);
        } catch (SQLException e) {
            log.warn("Failed to clean orphan locks. {}", e.getMessage());
            log.debug("Failed to clean orphan locks.", e);
        }
    }

    @Override
    public void init() {
        String serverId = ContextHelper.get().getServerId();
        try {
            int effectedRows = dbDistributeLocksDao.deleteAllOwnerLocks(serverId);
            log.debug("{} rows were removed during initialization.", effectedRows);
        } catch (SQLException e) {
            log.warn("Failed deleting old locks during initialization. {}", e.getMessage());
            log.debug("Failed deleting old locks during initialization.", e);
        }
    }

    @Override
    public void reload(CentralConfigDescriptor oldDescriptor, List<DataDiff<?>> configDiff) {

    }

    @Override
    public void cleanDbExpiredLocks() {
        log.debug("Attempting to release DB expired locks.");
        Set<LockInfo> locksToClean = Sets.newHashSet();
        try {
            long current = System.currentTimeMillis();
            log.trace("Current time is: {}", current);
            long minAllowedTime = getMinAllowedTime(current, hazelcastMaxLockLeaseTime.getLong());
            Set<LockInfo> expiredLocks = dbDistributeLocksDao.getExpiredLocks(minAllowedTime).stream()
                    .filter(filterOutActiveReplicationLocks(current))
                    .collect(Collectors.toSet());
            locksToClean.addAll(expiredLocks);
        } catch (SQLException e) {
            throw new StorageException("Failed to list locks.", e);
        }
        forceRemoveLocks(locksToClean, false);
    }

    @Override
    public void cleanCachedExpiredLocks() {
        log.debug("Attempting to release cached expired locks.");
        long current = System.currentTimeMillis();
        log.trace("Current time is: {}", current);
        long minAllowedTime = getMinAllowedTime(current, hazelcastMaxLockLeaseTime.getLong());
        Set<LockInfo> locks = localLocks.values().stream()
                .filter(Objects::nonNull)
                .filter(lockInfo -> lockInfo.getStartedTime() < minAllowedTime)
                .filter(filterOutActiveReplicationLocks(current))
                .collect(Collectors.toSet());
        if (CollectionUtils.notNullOrEmpty(locks)) {
            log.debug("Cleaning cached '{}' expired locks", locks.size());
            forceRemoveLocks(locks, true);
        }
    }

    @Override
    public void destroy() {
        String owner = ContextHelper.get().getServerId();
        log.debug("Destroying all locks for server: {} ", owner);
        try {
            Set<LockInfo> locks = dbDistributeLocksDao.getAllCurrentServerLocks(owner);
            forceRemoveLocks(locks, false);
        } catch (SQLException e) {
            throw new RuntimeException("" +
                    "Failed to destroy locks for server: " + owner);
        }
    }

    /**
     * Replication locks lease time out is 24x bigger than the regular lock lease time, therefore, we filter out all
     * the replication locks that their expiry is less than 24x then the regular lease time.
     */
    private Predicate<LockInfo> filterOutActiveReplicationLocks(long current) {
        return lockInfo -> {
            if (isReplicationLock(lockInfo)) {
                long minAllowedTimeForReplication = getMinAllowedTime(current,
                        hazelcastMaxLockLeaseTime.getLong() * 24);
                return lockInfo.getStartedTime() < minAllowedTimeForReplication;
            } else {
                return true;
            }
        };
    }

    private boolean isReplicationLock(LockInfo lockInfo) {
        return REPLICATION_LOCK_CATEGORY.equals(lockInfo.getCategory());
    }

    private long getMinAllowedTime(long current, long maxLeaseTimeMinutes) {
        return current - TimeUnit.MINUTES.toMillis(maxLeaseTimeMinutes);
    }

    /**
     * @param locks            List of locks to release, should not be null or empty.
     * @param removeCachedOnly Whether to release the locks from the cache only, or from both DB and cache
     */
    private void forceRemoveLocks(@Nonnull Set<LockInfo> locks, boolean removeCachedOnly) {
        log.debug("Cleaning '{}' expired locks", locks.size());
        for (LockInfo lock : locks) {
            log.debug("Destroying lock for category: {}, key: {} ", lock.getCategory(), lock.getKey());
            localLocks.remove(toLocalId(lock.getCategory(), lock.getKey()));
            if (!removeCachedOnly) {
                try {
                    DbUnlockSupplier unlockSupplier = getForceUnlockSupplier(lock.getCategory(), lock.getKey());
                    unlockInternal(unlockSupplier, lock.getCategory(), lock.getKey(), null, 1);
                } catch (Exception e) {
                    // the lock might have already deleted, so we ignore.
                    log.debug("Failed to destroy lock for category: {}, key: {} . {}", lock.getCategory(),
                            lock.getKey(), e.getMessage());
                    log.trace("Failed to destroy lock.", e);
                }
            }
        }
    }

    private DbUnlockSupplier getForceUnlockSupplier(String category, String key) {
        return () -> dbDistributeLocksDao.releaseForceLock(category, key);
    }

    private DbUnlockSupplier buildUnlockSupplier(String category, String key, String owner) {
        return () -> dbDistributeLocksDao.deleteLock(category, key, owner);
    }

    public String toLocalId(String category, String key) {
        return "category:" + category + ",key:" + key;
    }

    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {

    }

}
