package org.artifactory.storage.db.search.itest;

import ch.qos.logback.classic.Level;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.ItemInfo;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.repo.RepoPath;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.artifactory.storage.db.search.NodeSearcher;
import org.artifactory.test.TestUtils;
import org.jfrog.storage.JdbcHelper;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import javax.inject.Inject;
import java.util.Set;
import java.util.stream.Collectors;

import static org.fest.assertions.Assertions.assertThat;

/**
 * @author Yossi Shaul
 */
@Test
public class NodeSearcherTest extends DbBaseTest {

    @Inject
    private NodeSearcher nodeSearcher;

    @BeforeClass
    public void setup() {
        importSql("/sql/nodes-for-searches.sql");
        TestUtils.setLoggingLevel(JdbcHelper.class, Level.INFO);
    }

    @Test
    public void findFilesByNameUnderPath() {
        String fileName = "sha256__0e03bdcc26d7a9a57ef3b6f1bf1a210cff6239bff7c8cac72435984032851689";
        Set<FileInfo> nodes = nodeSearcher.findFiles("docker", "hello", fileName);
        assertThat(toRepoPaths(nodes)).containsOnly(
                new RepoPathImpl("docker", "hello/latest/" + fileName),
                new RepoPathImpl("docker", "hello/1.0/" + fileName)
        );
    }

    @Test
    public void findFilesByNameFromRootPath() {
        Set<FileInfo> nodes = nodeSearcher.findFiles("docker", "", "manifest.json");
        assertThat(toRepoPaths(nodes)).containsOnly(
                new RepoPathImpl("docker", "hello/latest/manifest.json"),
                new RepoPathImpl("docker", "hello/1.0/manifest.json"),
                new RepoPathImpl("docker", "ubuntu/latest/manifest.json")
        );
    }

    @Test
    public void findFilesByNameDirectlyUnderRootPath() {
        Set<FileInfo> nodes = nodeSearcher.findFiles("repo1", "", "readme.md");
        assertThat(toRepoPaths(nodes)).containsOnly(new RepoPathImpl("repo1", "readme.md"));
    }

    @Test
    public void findFilesBySha256UnderPath() {
        String sha256 = "0e03bdcc26d7a9a57ef3b6f1bf1a210cff6239bff7c8cac72435984032851689";
        Set<FileInfo> nodes = nodeSearcher.findFilesBySha256("docker", "hello", sha256);
        String fieName = "sha256__" + sha256;
        assertThat(toRepoPaths(nodes)).containsOnly(
                new RepoPathImpl("docker", "hello/latest/" + fieName),
                new RepoPathImpl("docker", "hello/1.0/" + fieName)
        );
    }

    @Test
    public void findFilesBySha256FromRootPath() {
        Set<FileInfo> nodes = nodeSearcher.findFilesBySha256("repo1", "",
                "0e03bdcc26d7a9a57ef3b6f1bf1a210cff6239bff7c8cac72435984032851670");
        assertThat(toRepoPaths(nodes)).containsOnly(
                new RepoPathImpl("repo1", "ant/ant/1.5/ant-1.5.jar")
        );
    }

    @Test
    public void findFilesBySha256DirectlyUnderRootPath() {
        Set<FileInfo> nodes = nodeSearcher
                .findFilesBySha256("repo1", "", "0e03bdcc26d7a9a57ef3b6f1bf1a210cff6239bff7c8cac72435984032851671");
        assertThat(toRepoPaths(nodes)).containsOnly(new RepoPathImpl("repo1", "readme.md"));
    }

    @Test
    public void findFilesBySha256UnderPathUsingPropertySearch() {
        setSha256Ready(false);  // force search by property
        String sha256 = "692c352adcf2821d6988021248da6b276cb738808f69dcc7bbb74a9c952146f7";
        Set<FileInfo> nodes = nodeSearcher.findFilesBySha256("docker", "ubuntu", sha256);
        String fieName = "sha256__" + sha256;
        assertThat(toRepoPaths(nodes)).containsOnly(
                new RepoPathImpl("docker", "ubuntu/latest/" + fieName)
        );
        setSha256Ready(true);
    }

    private void setSha256Ready(boolean ready) {
        InternalDbService dbService = ContextHelper.get().beanForType(InternalDbService.class);
        TestUtils.setField(dbService, "sha256Ready", ready);
    }

    private Set<RepoPath> toRepoPaths(Set<FileInfo> nodes) {
        return nodes.stream().map(ItemInfo::getRepoPath).collect(Collectors.toSet());
    }
}