package org.artifactory.webapp.servlet.util;

import org.apache.commons.lang.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.util.HttpUtils;
import org.artifactory.webapp.servlet.authentication.ArtifactoryAccessTokenAuthenticationFilter;
import org.artifactory.webapp.servlet.authentication.ArtifactoryAuthenticationFilter;
import org.artifactory.webapp.servlet.authentication.ArtifactoryBasicAuthenticationFilter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.authentication.BadCredentialsException;

import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.constraints.NotNull;
import java.util.Arrays;

import static org.apache.http.protocol.HTTP.SERVER_HEADER;


/**
 * @author Mor Merhav
 */
public final class ServletUtils {

    private static final Logger log = LoggerFactory.getLogger(ServletUtils.class);

    private ServletUtils() {

    }

    /**
     * @return login identifier
     */
    public static String getLoginIdentifier(ServletRequest servletRequest, ArtifactoryAuthenticationFilter filter) {
        String loginIdentifier = filter.getCacheKey(servletRequest);
        try {
            // fetch context LoginIdentifier
            loginIdentifier = filter.getLoginIdentifier(servletRequest);
            if (filter instanceof ArtifactoryAccessTokenAuthenticationFilter ||
                    filter instanceof ArtifactoryBasicAuthenticationFilter) {
                log.debug("Found loginIdentifier {}", loginIdentifier);
            }
        } catch (BadCredentialsException e) {
            log.debug("Resolving uses access details has failed, {}", e.getMessage());
            if (loginIdentifier == null) { loginIdentifier = ""; }
        }
        return loginIdentifier;
    }

    public static void addSensitiveHeadersToResponse(ServletRequest request, ServletResponse response,
            ArtifactoryContext artifactoryContext) {
        if (response instanceof HttpServletResponse) {
            HttpServletResponse httpResponse = (HttpServletResponse) response;
            String requestURI = ((HttpServletRequest) request).getRequestURI();
            if (isBinaryStoreServletRequest(requestURI)) {
                addBinStoreHeaders(httpResponse);
            } else {
                addArtifactoryHeaders(httpResponse, artifactoryContext);
            }
        }
    }

    public static boolean isBinaryStoreServletRequest(@NotNull String requestURI) {
        String[] uriParts = requestURI.split("/");
        return Arrays.asList(uriParts).contains("binarystore");
    }

    private static void addArtifactoryHeaders(HttpServletResponse httpResponse, ArtifactoryContext artifactoryContext) {
        if (!httpResponse.containsHeader(SERVER_HEADER)) {
            //Add the server header (curl -I http://localhost:8080/artifactory/)
            httpResponse.setHeader(SERVER_HEADER, HttpUtils.getArtifactoryUserAgent());
        }

        // set the Artifactory instance id header
        String hostId =
                artifactoryContext.beanForType(AddonsManager.class).addonByType(HaCommonAddon.class)
                        .getHostId();
        httpResponse.setHeader(ArtifactoryResponse.ARTIFACTORY_ID, hostId);

        String serverId = artifactoryContext.getServerId();
        if (StringUtils.isNotBlank(serverId) && !HaCommonAddon.ARTIFACTORY_PRO.equals(serverId)) {
            httpResponse.setHeader(HaCommonAddon.ARTIFACTORY_NODE_ID, serverId);
        }
    }

    private static void addBinStoreHeaders(HttpServletResponse httpResponse) {
        if (!httpResponse.containsHeader(SERVER_HEADER)) {
            httpResponse.setHeader(SERVER_HEADER, "BinaryStore/1.0");
        }
    }
}
